@@ -24,30 +24,36 @@ When and if the MyPy issue is resolved, the overloads will be removed.
 """
 import inspect
 from typing import (
-    Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Sequence, Tuple,
-    Type, TypeVar, cast, overload,
+    Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Tuple, Type,
+    TypeVar, Union, cast, overload,
 )
 
 import click
-from click import HelpFormatter
 
+import cloup
 from ._context import Context
 from ._option_groups import OptionGroupMixin
 from ._sections import Section, SectionMixin
-from ._util import first_bool, reindent, click_version_ge_8_1
+from ._util import click_version_ge_8_1, first_bool, reindent
 from .constraints import ConstraintMixin
+from .typing import AnyCallable
 
 ClickCommand = TypeVar('ClickCommand', bound=click.Command)
 ClickGroup = TypeVar('ClickGroup', bound=click.Group)
 
 
-class BaseCommand(click.Command):
-    """Base class for cloup commands.
+class Command(ConstraintMixin, OptionGroupMixin, click.Command):
+    """A ``click.Command`` supporting option groups and constraints.
 
-    * It back-ports a feature from Click v8.0, i.e. the ``context_class``
-      class attribute, which is set to ``cloup.Context``.
+    Refer to superclasses for the documentation of all accepted parameters:
 
-    * It adds a ``formatter_settings`` instance attribute.
+    - :class:`ConstraintMixin`
+    - :class:`OptionGroupMixin`
+    - :class:`click.Command`
+
+    Besides other things, this class also:
+
+    * adds a ``formatter_settings`` instance attribute.
 
     Refer to :class:`click.Command` for the documentation of all parameters.
 
@@ -56,10 +62,10 @@ class BaseCommand(click.Command):
     context_class: Type[Context] = Context
 
     def __init__(
-        self, *args,
+        self, *args: Any,
         aliases: Optional[Iterable[str]] = None,
         formatter_settings: Optional[Dict[str, Any]] = None,
-        **kwargs,
+        **kwargs: Any,
     ):
         super().__init__(*args, **kwargs)
         #: HelpFormatter options that are merged with ``Context.formatter_settings``
@@ -68,83 +74,63 @@ class BaseCommand(click.Command):
         self.formatter_settings: Dict[str, Any] = (
             {} if formatter_settings is None else formatter_settings)
 
-    def make_context(self, info_name, args, parent=None, **extra) -> Context:
-        for key, value in self.context_settings.items():
-            if key not in extra:
-                extra[key] = value
-
-        ctx = self.context_class(self, info_name=info_name, parent=parent, **extra)
-
-        with ctx.scope(cleanup=False):
-            self.parse_args(ctx, args)
-        return ctx
-
     def get_normalized_epilog(self) -> str:
         if self.epilog and click_version_ge_8_1:
             return inspect.cleandoc(self.epilog)
         return self.epilog or ""
 
     # Differently from Click, this doesn't indent the epilog.
-    def format_epilog(self, ctx, formatter):
+    def format_epilog(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
         if self.epilog:
+            assert isinstance(formatter, cloup.HelpFormatter)
             epilog = self.get_normalized_epilog()
             formatter.write_paragraph()
             formatter.write_epilog(epilog)
 
-    def format_help_text(self, ctx, formatter):
+    def format_help_text(
+        self, ctx: click.Context, formatter: click.HelpFormatter
+    ) -> None:
+        assert isinstance(formatter, cloup.HelpFormatter)
         formatter.write_command_help_text(self)
 
-    def format_aliases(self, ctx, formatter):
+    def format_aliases(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
         if not self.aliases:
             return
+        assert isinstance(formatter, cloup.HelpFormatter)
         formatter.write_aliases(self.aliases)
 
-    def format_help(self, ctx: click.Context, formatter: HelpFormatter) -> None:
+    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
         self.format_usage(ctx, formatter)
         self.format_aliases(ctx, formatter)
         self.format_help_text(ctx, formatter)
-        self.format_options(ctx, formatter)
+        self.format_params(ctx, formatter)
+        if self.must_show_constraints(ctx):
+            self.format_constraints(ctx, formatter)  # type: ignore
+        if isinstance(self, click.MultiCommand):
+            self.format_commands(ctx, formatter)
         self.format_epilog(ctx, formatter)
 
 
-class Command(ConstraintMixin, OptionGroupMixin, BaseCommand):
-    """
-    A ``click.Command`` supporting option groups and constraints.
-
-    Refer to superclasses for the documentation of all accepted parameters:
-
-    - :class:`ConstraintMixin`
-    - :class:`OptionGroupMixin`
-    - :class:`BaseCommand` -> :class:`click.Command`
-
-    .. versionchanged:: 0.8.0
-        this class now inherits from :class:`cloup.BaseCommand`.
-    """
-    pass
-
-
-class Group(SectionMixin, BaseCommand, click.Group):
+class Group(SectionMixin, Command, click.Group):
     """
     A ``click.Group`` that allows to organize its subcommands in multiple help
-    sections and and whose subcommands are, by default, of type
-    :class:`cloup.Command`.
-
-    This class is just a :class:`click.Group` mixed with :class:`SectionMixin`
-    that overrides the decorators :meth:`command` and :meth:`group` so that
-    a ``section`` for the created subcommand can be specified.
+    sections and whose subcommands are, by default, of type :class:`cloup.Command`.
 
     Refer to superclasses for the documentation of all accepted parameters:
 
     - :class:`SectionMixin`
-    - :class:`BaseCommand` -> :class:`click.Command`
+    - :class:`Command`
     - :class:`click.Group`
 
-    This class adds a single parameter:
+    Apart from superclasses arguments, the following is the only additional parameter:
 
     ``show_subcommand_aliases``: ``Optional[bool] = None``
         whether to show subcommand aliases; aliases are shown by default and
         can be disabled using this argument or the homonym context setting.
 
+    .. versionchanged:: 0.14.0
+        this class now supports option groups and constraints.
+
     .. versionadded:: 0.10.0
         the "command aliases" feature, including the ``show_subcommand_aliases``
         parameter/attribute.
@@ -154,10 +140,12 @@ class Group(SectionMixin, BaseCommand, click.Group):
     """
     SHOW_SUBCOMMAND_ALIASES: bool = False
 
-    def __init__(self, *args, show_subcommand_aliases: Optional[bool] = None, **kwargs):
+    def __init__(
+        self, *args: Any, show_subcommand_aliases: Optional[bool] = None, **kwargs: Any
+    ):
         super().__init__(*args, **kwargs)
         self.show_subcommand_aliases = show_subcommand_aliases
-        """Whether to show subcommand aliases. This """
+        """Whether to show subcommand aliases."""
 
         self.alias2name: Dict[str, str] = {}
         """Dictionary mapping each alias to a command name."""
@@ -265,10 +253,11 @@ class Group(SectionMixin, BaseCommand, click.Group):
         deprecated: bool = False,
         align_option_groups: Optional[bool] = None,
         show_constraints: Optional[bool] = None,
-    ) -> Callable[[Callable], Command]:
+        params: Optional[List[click.Parameter]] = None,
+    ) -> Callable[[AnyCallable], Command]:
         ...
 
-    @overload  # type: ignore
+    @overload
     def command(  # Why overloading? Refer to module docstring.
         self, name: Optional[str] = None,
         *,
@@ -284,11 +273,18 @@ class Group(SectionMixin, BaseCommand, click.Group):
         no_args_is_help: bool = False,
         hidden: bool = False,
         deprecated: bool = False,
-        **kwargs,
-    ) -> Callable[[Callable], ClickCommand]:
+        params: Optional[List[click.Parameter]] = None,
+        **kwargs: Any,
+    ) -> Callable[[AnyCallable], ClickCommand]:
         ...
 
-    def command(self, name=None, *, cls=None, section=None, **kwargs):
+    def command(
+        self, name: Optional[str] = None, *,
+        aliases: Optional[Iterable[str]] = None,
+        cls: Optional[Type[ClickCommand]] = None,
+        section: Optional[Section] = None,
+        **kwargs: Any
+    ) -> Callable[[AnyCallable], Union[Command, ClickCommand]]:
         """Returns a decorator that creates a new subcommand of this ``Group``
         using the decorated function as callback.
 
@@ -300,9 +296,10 @@ class Group(SectionMixin, BaseCommand, click.Group):
         .. versionchanged:: 0.10.0
             all arguments but ``name`` are now keyword-only.
         """
+        make_command = command(name=name, cls=cls, aliases=aliases, **kwargs)
 
-        def decorator(f):
-            cmd = command(name=name, cls=cls, **kwargs)(f)
+        def decorator(f: AnyCallable) -> Union[Command, ClickCommand]:
+            cmd = make_command(f)
             self.add_command(cmd, section=section)
             return cmd
 
@@ -333,7 +330,7 @@ class Group(SectionMixin, BaseCommand, click.Group):
         chain: bool = False,
         hidden: bool = False,
         deprecated: bool = False,
-    ) -> Callable[[Callable], 'Group']:
+    ) -> Callable[[AnyCallable], 'Group']:
         ...
 
     @overload
@@ -354,11 +351,19 @@ class Group(SectionMixin, BaseCommand, click.Group):
         chain: bool = False,
         hidden: bool = False,
         deprecated: bool = False,
-        **kwargs
-    ) -> Callable[[Callable], ClickGroup]:
+        params: Optional[List[click.Parameter]] = None,
+        **kwargs: Any
+    ) -> Callable[[AnyCallable], ClickGroup]:
         ...
 
-    def group(self, name=None, *, cls=None, section=None, **kwargs):
+    def group(  # type: ignore
+        self, name: Optional[None] = None,
+        *,
+        cls: Optional[Type[ClickGroup]] = None,
+        aliases: Optional[Iterable[str]] = None,
+        section: Optional[Section] = None,
+        **kwargs: Any
+    ) -> Callable[[AnyCallable], Union["Group", ClickGroup]]:
         """Returns a decorator that creates a new subcommand of this ``Group``
         using the decorated function as callback.
 
@@ -370,9 +375,10 @@ class Group(SectionMixin, BaseCommand, click.Group):
         .. versionchanged:: 0.10.0
             all arguments but ``name`` are now keyword-only.
         """
+        make_group = group(name=name, cls=cls, aliases=aliases, **kwargs)
 
-        def decorator(f):
-            cmd = group(name=name, cls=cls, **kwargs)(f)
+        def decorator(f: AnyCallable) -> Union["Group", ClickGroup]:
+            cmd = make_group(f)
             self.add_command(cmd, section=section)
             return cmd
 
@@ -397,8 +403,9 @@ def command(
     hidden: bool = False,
     deprecated: bool = False,
     align_option_groups: Optional[bool] = None,
-    show_constraints: Optional[bool] = None
-) -> Callable[[Callable], Command]:
+    show_constraints: Optional[bool] = None,
+    params: Optional[List[click.Parameter]] = None,
+) -> Callable[[AnyCallable], Command]:
     ...
 
 
@@ -417,13 +424,19 @@ def command(  # In this overload: "cls: ClickCommand"
     no_args_is_help: bool = False,
     hidden: bool = False,
     deprecated: bool = False,
-    **kwargs
-) -> Callable[[Callable], ClickCommand]:
+    params: Optional[List[click.Parameter]] = None,
+    **kwargs: Any
+) -> Callable[[AnyCallable], ClickCommand]:
     ...
 
 
 # noinspection PyIncorrectDocstring
-def command(name=None, *, aliases=None, cls=None, **kwargs):
+def command(
+    name: Optional[str] = None, *,
+    aliases: Optional[Iterable[str]] = None,
+    cls: Optional[Type[ClickCommand]] = None,
+    **kwargs: Any
+) -> Callable[[AnyCallable], Union[Command, ClickCommand]]:
     """
     Returns a decorator that creates a new command using the decorated function
     as callback.
@@ -438,7 +451,7 @@ def command(name=None, *, aliases=None, cls=None, **kwargs):
 
     Note that the following arguments are about Cloup-specific features and are
     not supported by all ``click.Command``, so if you provide a custom ``cls``
-    make sure you don't ne:
+    make sure you don't set these:
 
     - ``formatter_settings``
     - ``align_option_groups`` (``cls`` needs to inherit from ``OptionGroupMixin``)
@@ -495,28 +508,37 @@ def command(name=None, *, aliases=None, cls=None, **kwargs):
         whether to include a "Constraint" section in the command help. This
         is also available as a context setting having a lower priority than
         this attribute.
+    :param params:
+        **(click >= 8.1.0)** a list of parameters (:class:`Argument` and
+        :class:`Option` instances). Params added with ``@option`` and ``@argument``
+        are appended to the end of the list if given.
     :param kwargs:
         any other argument accepted by the instantiated command class (``cls``).
     """
+    if callable(name):
+        raise Exception(
+            f"you forgot parenthesis in the command decorator for `{name.__name__}`. "
+            f"While parenthesis are optional in Click > 8.1, they are required in Cloup."
+        )
 
-    def decorator(f):
+    def decorator(f: AnyCallable) -> ClickCommand:
         if hasattr(f, '__constraints'):
             if cls and not issubclass(cls, ConstraintMixin):
                 raise TypeError(
                     f"a Command must inherit from cloup.ConstraintMixin to support "
                     f"constraints; {cls} doesn't")
-            constraints = tuple(reversed(f.__constraints))
-            del f.__constraints
+            constraints = tuple(reversed(f.__constraints))  # type: ignore
+            del f.__constraints  # type: ignore
             kwargs['constraints'] = constraints
 
         cmd_cls = cls if cls is not None else Command
         try:
-            cmd = click.command(name, cls=cmd_cls, **kwargs)(f)
+            cmd = cast(ClickCommand, click.command(name, cls=cmd_cls, **kwargs)(f))
             if aliases:
-                cmd.aliases = list(aliases)
+                cmd.aliases = list(aliases)  # type: ignore
             return cmd
         except TypeError as error:
-            raise _process_unexpected_kwarg_error(error, _ARGS_INFO, cls)
+            raise _process_unexpected_kwarg_error(error, _ARGS_INFO, cmd_cls)
 
     return decorator
 
@@ -525,8 +547,8 @@ def command(name=None, *, aliases=None, cls=None, **kwargs):
 def group(
     name: Optional[str] = None,
     *,
-    aliases: Optional[Iterable[str]] = None,
     cls: None = None,
+    aliases: Optional[Iterable[str]] = None,
     sections: Iterable[Section] = (),
     align_sections: Optional[bool] = None,
     invoke_without_command: bool = False,
@@ -542,7 +564,8 @@ def group(
     chain: bool = False,
     hidden: bool = False,
     deprecated: bool = False,
-) -> Callable[[Callable], Group]:
+    params: Optional[List[click.Parameter]] = None,
+) -> Callable[[AnyCallable], Group]:
     ...
 
 
@@ -550,8 +573,8 @@ def group(
 def group(
     name: Optional[str] = None,
     *,
-    aliases: Optional[Iterable[str]] = None,
     cls: Type[ClickGroup],
+    aliases: Optional[Iterable[str]] = None,
     invoke_without_command: bool = False,
     no_args_is_help: bool = False,
     context_settings: Optional[Dict[str, Any]] = None,
@@ -564,12 +587,15 @@ def group(
     chain: bool = False,
     hidden: bool = False,
     deprecated: bool = False,
-    **kwargs
-) -> Callable[[Callable], ClickGroup]:
+    params: Optional[List[click.Parameter]] = None,
+    **kwargs: Any
+) -> Callable[[AnyCallable], ClickGroup]:
     ...
 
 
-def group(name=None, *, cls=None, **kwargs):
+def group(
+    name: Optional[str] = None, *, cls: Optional[Type[ClickGroup]] = None, **kwargs: Any
+) -> Callable[[AnyCallable], click.Group]:
     """
     Returns a decorator that instantiates a ``Group`` (or a subclass of it)
     using the decorated function as callback.
@@ -632,41 +658,45 @@ def group(name=None, *, cls=None, **kwargs):
     :param subcommand_metavar:
         string used in the command's usage string to indicate the subcommand place.
     :param chain:
-        if this is set to `True` chaining of multiple subcommands is enabled.
+        if this is set to `True`, chaining of multiple subcommands is enabled.
         This restricts the form of commands in that they cannot have optional
         arguments but it allows multiple commands to be chained together.
+    :param params:
+        **(click >= 8.1.0)** a list of parameters (:class:`Argument` and
+        :class:`Option` instances). Params added with ``@option`` and ``@argument``
+        are appended to the end of the list if given.
     :param kwargs:
         any other argument accepted by the instantiated command class.
     """
     if cls is None:
-        cls = Group
-    elif not issubclass(cls, click.Group):
+        return command(name=name, cls=Group, **kwargs)
+    elif issubclass(cls, click.Group):
+        return command(name=name, cls=cls, **kwargs)
+    else:
         raise TypeError(
             'this decorator requires cls to be a click.Group (or a subclass).')
 
-    return command(name=name, cls=cls, **kwargs)
-
 
 # Side stuff for better error messages
 
 class _ArgInfo(NamedTuple):
     arg_name: str
-    requires: Type
-    supported_by: Sequence[str] = []
+    requires: Type[Any]
+    supported_by: str = ""
 
 
 _ARGS_INFO = {
     info.arg_name: info for info in [
-        _ArgInfo('formatter_settings', BaseCommand, ['cloup.Command', 'cloup.Group']),
-        _ArgInfo('align_option_groups', OptionGroupMixin, ['cloup.Command']),
-        _ArgInfo('show_constraints', ConstraintMixin, ['cloup.Command']),
-        _ArgInfo('align_sections', SectionMixin, ['cloup.Group'])
+        _ArgInfo('formatter_settings', Command, "both Command and Group"),
+        _ArgInfo('align_option_groups', OptionGroupMixin, "both Command and Group"),
+        _ArgInfo('show_constraints', ConstraintMixin, "both Command and Group"),
+        _ArgInfo('align_sections', SectionMixin, "Group")
     ]
 }
 
 
 def _process_unexpected_kwarg_error(
-    error: TypeError, args_info: Dict[str, _ArgInfo], cls: Type[click.Command]
+    error: TypeError, args_info: Dict[str, _ArgInfo], cls: Type[Command]
 ) -> TypeError:
     """Checks if the developer tried to pass a Cloup-specific argument to a ``cls``
     that doesn't support it and if that's the case, augments the error message
@@ -680,10 +710,9 @@ def _process_unexpected_kwarg_error(
     arg = match.group()
     info = args_info[arg]
     extra_info = reindent(f"""\n
-        HINT: you set cls={cls} but this class
-        doesn't support the argument "{arg}".
-        In Cloup, this argument is handled by {info.requires.__name__} and
-        it's supported by {", ".join(info.supported_by)}.
+        HINT: you set cls={cls} but this class doesn't support the argument "{arg}".
+        In Cloup, this argument is supported by {info.supported_by}
+        via {info.requires.__name__}.
     """, 4)
     new_message = message + '\n' + extra_info
     return TypeError(new_message)
